@@ -22,6 +22,7 @@
 #include "gspell-entry-buffer.h"
 #include "gspell-entry-utils.h"
 #include "gspell-context-menu.h"
+#include "gspell-current-word-policy.h"
 
 /**
  * SECTION:entry
@@ -54,6 +55,8 @@ struct _GspellEntry
 	GtkEntryBuffer *buffer;
 	GspellChecker *checker;
 
+	GspellCurrentWordPolicy *current_word_policy;
+
 	/* List elements: GspellEntryWord*.
 	 * Used for unit tests.
 	 */
@@ -79,19 +82,40 @@ enum
 
 G_DEFINE_TYPE (GspellEntry, gspell_entry, G_TYPE_OBJECT)
 
-static GspellChecker *
-get_checker (GspellEntry *gspell_entry)
+/* This function should be called instead of accessing the inline_spell_checking
+ * attribute.
+ */
+static gboolean
+inline_spell_checking_is_enabled (GspellEntry *gspell_entry)
 {
-	GspellEntryBuffer *gspell_buffer;
+	/* The GtkEntry:input-purpose and/or GtkEntry:input-hints could be taken
+	 * into account here, but it is not the case. There is already the
+	 * GspellEntry:inline-spell-checking property, which needs to be FALSE
+	 * by default. If it was TRUE by default, an application would just need
+	 * to call gspell_entry_get_from_gtk_entry(), but it would be strange to
+	 * do nothing with the returned GspellEntry. So inline-spell-checking is
+	 * FALSE by default and the application anyway needs to set it to TRUE
+	 * manually to enable the *inline* spell checking (a GtkEntry could have
+	 * other types of spell checking, for example based on GspellNavigator
+	 * to check an entire form or check a list of forms, even though such
+	 * feature is probably rare).
+	 *
+	 * In other words, it might be desirable to set
+	 * GTK_INPUT_HINT_SPELLCHECK but keeping the inline spell checking of
+	 * GspellEntry disabled. But when the inline spell checker of
+	 * GspellEntry is enabled, it is normally always desirable to set
+	 * GTK_INPUT_HINT_SPELLCHECK, which can be seen as duplicated state, but
+	 * it is not, because if the GspellEntry:inline-spell-checking property
+	 * is removed, another boolean property would be needed to tell
+	 * GspellEntry whether it needs to bind the input-hints settings to its
+	 * inline spell checker.
+	 *
+	 * Anyway, the mere fact of calling gspell_entry_get_from_gtk_entry()
+	 * should not have unexpected side effects.
+	 */
 
-	if (gspell_entry->buffer == NULL)
-	{
-		return NULL;
-	}
-
-	gspell_buffer = gspell_entry_buffer_get_from_gtk_entry_buffer (gspell_entry->buffer);
-
-	return gspell_entry_buffer_get_spell_checker (gspell_buffer);
+	return (gspell_entry->inline_spell_checking &&
+		gtk_entry_get_visibility (gspell_entry->entry));
 }
 
 static void
@@ -185,20 +209,18 @@ insert_underline (GspellEntry *gspell_entry,
 static void
 update_misspelled_words_list (GspellEntry *gspell_entry)
 {
-	GspellChecker *checker;
 	GSList *all_words;
 
 	g_slist_free_full (gspell_entry->misspelled_words, _gspell_entry_word_free);
 	gspell_entry->misspelled_words = NULL;
 
-	if (!gspell_entry->inline_spell_checking)
+	if (!inline_spell_checking_is_enabled (gspell_entry))
 	{
 		return;
 	}
 
-	checker = get_checker (gspell_entry);
-	if (checker == NULL ||
-	    gspell_checker_get_language (checker) == NULL)
+	if (gspell_entry->checker == NULL ||
+	    gspell_checker_get_language (gspell_entry->checker) == NULL)
 	{
 		return;
 	}
@@ -211,7 +233,7 @@ update_misspelled_words_list (GspellEntry *gspell_entry)
 		gboolean correctly_spelled;
 		GError *error = NULL;
 
-		correctly_spelled = gspell_checker_check_word (checker,
+		correctly_spelled = gspell_checker_check_word (gspell_entry->checker,
 							       cur_word->word_str, -1,
 							       &error);
 
@@ -242,6 +264,17 @@ update_misspelled_words_list (GspellEntry *gspell_entry)
 	gspell_entry->misspelled_words = g_slist_reverse (gspell_entry->misspelled_words);
 }
 
+static gboolean
+is_current_word (GspellEntry     *gspell_entry,
+		 GspellEntryWord *word)
+{
+	gint cursor_pos;
+
+	cursor_pos = gtk_editable_get_position (GTK_EDITABLE (gspell_entry->entry));
+
+	return (word->char_start <= cursor_pos && cursor_pos <= word->char_end);
+}
+
 static void
 recheck_all (GspellEntry *gspell_entry)
 {
@@ -253,6 +286,12 @@ recheck_all (GspellEntry *gspell_entry)
 	{
 		GspellEntryWord *cur_word = l->data;
 
+		if (!_gspell_current_word_policy_get_check_current_word (gspell_entry->current_word_policy) &&
+		    is_current_word (gspell_entry, cur_word))
+		{
+			continue;
+		}
+
 		insert_underline (gspell_entry,
 				  cur_word->byte_start,
 				  cur_word->byte_end);
@@ -322,6 +361,23 @@ notify_attributes_cb (GtkEntry    *gtk_entry,
 	}
 }
 
+static void
+language_notify_cb (GspellChecker *checker,
+		    GParamSpec    *pspec,
+		    GspellEntry   *gspell_entry)
+{
+	_gspell_current_word_policy_language_changed (gspell_entry->current_word_policy);
+	emit_changed_signal (gspell_entry);
+}
+
+static void
+session_cleared_cb (GspellChecker *checker,
+		    GspellEntry   *gspell_entry)
+{
+	_gspell_current_word_policy_session_cleared (gspell_entry->current_word_policy);
+	emit_changed_signal (gspell_entry);
+}
+
 static void
 set_checker (GspellEntry   *gspell_entry,
 	     GspellChecker *checker)
@@ -333,6 +389,14 @@ set_checker (GspellEntry   *gspell_entry,
 
 	if (gspell_entry->checker != NULL)
 	{
+		g_signal_handlers_disconnect_by_func (gspell_entry->checker,
+						      language_notify_cb,
+						      gspell_entry);
+
+		g_signal_handlers_disconnect_by_func (gspell_entry->checker,
+						      session_cleared_cb,
+						      gspell_entry);
+
 		g_signal_handlers_disconnect_by_func (gspell_entry->checker,
 						      emit_changed_signal,
 						      gspell_entry);
@@ -344,15 +408,15 @@ set_checker (GspellEntry   *gspell_entry,
 
 	if (gspell_entry->checker != NULL)
 	{
-		g_signal_connect_swapped (gspell_entry->checker,
-					  "notify::language",
-					  G_CALLBACK (emit_changed_signal),
-					  gspell_entry);
+		g_signal_connect (gspell_entry->checker,
+				  "notify::language",
+				  G_CALLBACK (language_notify_cb),
+				  gspell_entry);
 
-		g_signal_connect_swapped (gspell_entry->checker,
-					  "session-cleared",
-					  G_CALLBACK (emit_changed_signal),
-					  gspell_entry);
+		g_signal_connect (gspell_entry->checker,
+				  "session-cleared",
+				  G_CALLBACK (session_cleared_cb),
+				  gspell_entry);
 
 		g_signal_connect_swapped (gspell_entry->checker,
 					  "word-added-to-personal",
@@ -371,7 +435,17 @@ set_checker (GspellEntry   *gspell_entry,
 static void
 update_checker (GspellEntry *gspell_entry)
 {
-	set_checker (gspell_entry, get_checker (gspell_entry));
+	GspellChecker *checker = NULL;
+
+	if (gspell_entry->buffer != NULL)
+	{
+		GspellEntryBuffer *gspell_buffer;
+
+		gspell_buffer = gspell_entry_buffer_get_from_gtk_entry_buffer (gspell_entry->buffer);
+		checker = gspell_entry_buffer_get_spell_checker (gspell_buffer);
+	}
+
+	set_checker (gspell_entry, checker);
 }
 
 static void
@@ -380,9 +454,45 @@ notify_spell_checker_cb (GspellEntryBuffer *gspell_buffer,
 			 GspellEntry       *gspell_entry)
 {
 	update_checker (gspell_entry);
+
+	_gspell_current_word_policy_checker_changed (gspell_entry->current_word_policy);
 	emit_changed_signal (gspell_entry);
 }
 
+static void
+inserted_text_cb (GtkEntryBuffer *buffer,
+		  guint           position,
+		  gchar          *chars,
+		  guint           n_chars,
+		  GspellEntry    *gspell_entry)
+{
+	if (n_chars > 1)
+	{
+		_gspell_current_word_policy_several_chars_inserted (gspell_entry->current_word_policy);
+	}
+	else
+	{
+		gunichar ch;
+		gboolean empty_selection;
+		gint cursor_pos;
+		gboolean at_cursor_pos;
+
+		ch = g_utf8_get_char (chars);
+
+		empty_selection = !gtk_editable_get_selection_bounds (GTK_EDITABLE (gspell_entry->entry),
+								      NULL,
+								      NULL);
+
+		cursor_pos = gtk_editable_get_position (GTK_EDITABLE (gspell_entry->entry));
+		at_cursor_pos = cursor_pos == (gint)position;
+
+		_gspell_current_word_policy_single_char_inserted (gspell_entry->current_word_policy,
+								  ch,
+								  empty_selection,
+								  at_cursor_pos);
+	}
+}
+
 static void
 set_buffer (GspellEntry    *gspell_entry,
 	    GtkEntryBuffer *gtk_buffer)
@@ -402,6 +512,10 @@ set_buffer (GspellEntry    *gspell_entry,
 						      notify_spell_checker_cb,
 						      gspell_entry);
 
+		g_signal_handlers_disconnect_by_func (gspell_entry->buffer,
+						      inserted_text_cb,
+						      gspell_entry);
+
 		g_object_unref (gspell_entry->buffer);
 	}
 
@@ -416,6 +530,11 @@ set_buffer (GspellEntry    *gspell_entry,
 				  G_CALLBACK (notify_spell_checker_cb),
 				  gspell_entry);
 
+		g_signal_connect (gspell_entry->buffer,
+				  "inserted-text",
+				  G_CALLBACK (inserted_text_cb),
+				  gspell_entry);
+
 		g_object_ref (gspell_entry->buffer);
 	}
 
@@ -489,6 +608,9 @@ button_press_event_cb (GtkEntry       *gtk_entry,
 			_gspell_entry_utils_get_char_position_at_event (gtk_entry, event);
 	}
 
+	_gspell_current_word_policy_cursor_moved (gspell_entry->current_word_policy);
+	emit_changed_signal (gspell_entry);
+
 	return GDK_EVENT_PROPAGATE;
 }
 
@@ -497,16 +619,14 @@ language_activated_cb (const GspellLanguage *lang,
 		       gpointer              user_data)
 {
 	GspellEntry *gspell_entry;
-	GspellChecker *checker;
 
 	g_return_if_fail (GSPELL_IS_ENTRY (user_data));
 
 	gspell_entry = GSPELL_ENTRY (user_data);
 
-	checker = get_checker (gspell_entry);
-	if (checker != NULL)
+	if (gspell_entry->checker != NULL)
 	{
-		gspell_checker_set_language (checker, lang);
+		gspell_checker_set_language (gspell_entry->checker, lang);
 	}
 }
 
@@ -549,7 +669,6 @@ populate_popup_cb (GtkEntry    *gtk_entry,
 	GtkWidget *menu_item;
 	GtkMenuItem *lang_menu_item;
 	GtkMenuItem *suggestions_menu_item;
-	GspellChecker *checker;
 	const GspellLanguage *current_language;
 	GspellEntryWord *word;
 	gboolean correctly_spelled;
@@ -562,13 +681,12 @@ populate_popup_cb (GtkEntry    *gtk_entry,
 
 	menu = GTK_MENU (popup);
 
-	if (!gspell_entry->inline_spell_checking)
+	if (!inline_spell_checking_is_enabled (gspell_entry))
 	{
 		return;
 	}
 
-	checker = get_checker (gspell_entry);
-	if (checker == NULL)
+	if (gspell_entry->checker == NULL)
 	{
 		return;
 	}
@@ -579,7 +697,7 @@ populate_popup_cb (GtkEntry    *gtk_entry,
 	gtk_widget_show (menu_item);
 
 	/* Prepend language sub-menu */
-	current_language = gspell_checker_get_language (checker);
+	current_language = gspell_checker_get_language (gspell_entry->checker);
 	lang_menu_item = _gspell_context_menu_get_language_menu_item (current_language,
 								      language_activated_cb,
 								      gspell_entry);
@@ -595,7 +713,7 @@ populate_popup_cb (GtkEntry    *gtk_entry,
 		return;
 	}
 
-	correctly_spelled = gspell_checker_check_word (checker,
+	correctly_spelled = gspell_checker_check_word (gspell_entry->checker,
 						       word->word_str, -1,
 						       &error);
 
@@ -609,7 +727,7 @@ populate_popup_cb (GtkEntry    *gtk_entry,
 
 	if (!correctly_spelled)
 	{
-		suggestions_menu_item = _gspell_context_menu_get_suggestions_menu_item (checker,
+		suggestions_menu_item = _gspell_context_menu_get_suggestions_menu_item (gspell_entry->checker,
 											word->word_str,
 											suggestion_activated_cb,
 											gspell_entry);
@@ -621,6 +739,129 @@ populate_popup_cb (GtkEntry    *gtk_entry,
 	_gspell_entry_word_free (word);
 }
 
+static void
+move_cursor_cb (GspellEntry *gspell_entry)
+{
+	_gspell_current_word_policy_cursor_moved (gspell_entry->current_word_policy);
+	emit_changed_signal (gspell_entry);
+}
+
+static gboolean
+is_inside_word (const GSList *words,
+		gint          char_pos)
+{
+	const GSList *l;
+
+	for (l = words; l != NULL; l = l->next)
+	{
+		const GspellEntryWord *cur_word = l->data;
+
+		if (cur_word->char_start <= char_pos && char_pos < cur_word->char_end)
+		{
+			return TRUE;
+		}
+	}
+
+	return FALSE;
+}
+
+static gboolean
+ends_word (const GSList *words,
+	   gint          char_pos)
+{
+	const GSList *l;
+
+	for (l = words; l != NULL; l = l->next)
+	{
+		const GspellEntryWord *cur_word = l->data;
+
+		if (cur_word->char_end == char_pos)
+		{
+			return TRUE;
+		}
+	}
+
+	return FALSE;
+}
+
+static void
+delete_text_before_cb (GtkEditable *editable,
+		       gint         start_pos,
+		       gint         end_pos,
+		       GspellEntry *gspell_entry)
+{
+	gint real_start_pos;
+	gint real_end_pos;
+	gint cursor_pos;
+	GSList *words;
+	gboolean empty_selection;
+	gboolean spans_several_lines;
+	gboolean several_chars;
+	gboolean cursor_pos_at_start;
+	gboolean cursor_pos_at_end;
+	gboolean start_is_inside_word;
+	gboolean start_ends_word;
+	gboolean end_is_inside_word;
+	gboolean end_ends_word;
+
+	real_start_pos = start_pos;
+
+	if (end_pos < 0)
+	{
+		real_end_pos = gtk_entry_get_text_length (gspell_entry->entry);
+	}
+	else
+	{
+		real_end_pos = end_pos;
+	}
+
+	if (real_start_pos == real_end_pos)
+	{
+		return;
+	}
+
+	if (real_start_pos > real_end_pos)
+	{
+		gint real_start_pos_copy;
+
+		/* swap */
+		real_start_pos_copy = real_start_pos;
+		real_start_pos = real_end_pos;
+		real_end_pos = real_start_pos_copy;
+	}
+
+	g_assert_cmpint (real_start_pos, <, real_end_pos);
+
+	empty_selection = !gtk_editable_get_selection_bounds (editable, NULL, NULL);
+	spans_several_lines = FALSE;
+	several_chars = (real_end_pos - real_start_pos) > 1;
+
+	cursor_pos = gtk_editable_get_position (editable);
+	cursor_pos_at_start = cursor_pos == real_start_pos;
+	cursor_pos_at_end = cursor_pos == real_end_pos;
+
+	words = _gspell_entry_utils_get_words (gspell_entry->entry);
+
+	start_is_inside_word = is_inside_word (words, real_start_pos);
+	start_ends_word = ends_word (words, real_start_pos);
+
+	end_is_inside_word = is_inside_word (words, real_end_pos);
+	end_ends_word = ends_word (words, real_end_pos);
+
+	g_slist_free_full (words, _gspell_entry_word_free);
+
+	_gspell_current_word_policy_text_deleted (gspell_entry->current_word_policy,
+						  empty_selection,
+						  spans_several_lines,
+						  several_chars,
+						  cursor_pos_at_start,
+						  cursor_pos_at_end,
+						  start_is_inside_word,
+						  start_ends_word,
+						  end_is_inside_word,
+						  end_ends_word);
+}
+
 static void
 set_entry (GspellEntry *gspell_entry,
 	   GtkEntry    *gtk_entry)
@@ -670,6 +911,38 @@ set_entry (GspellEntry *gspell_entry,
 				G_CALLBACK (populate_popup_cb),
 				gspell_entry);
 
+	/* What we want here is to be notified when the cursor moved, but _not_
+	 * when the cursor moved because of a text insertion/deletion. To call
+	 * _gspell_current_word_policy_cursor_moved().
+	 *
+	 * Connecting to notify::cursor-position is not suitable because we have
+	 * notifications also when text is inserted/deleted. And we get the
+	 * notification *after* the GtkEditable::insert-text signal (not
+	 * *during* its emission). So it seems that the only simple solution is
+	 * to connect to ::move-cursor, even if its documentation doesn't
+	 * recommend that (normally, it should be used only to emit the signal).
+	 *
+	 * Note that _gspell_current_word_policy_cursor_moved() is also called
+	 * in button_press_event_cb().
+	 *
+	 * The GtkEntry API is not really convenient, if you find a better
+	 * solution, or if you improve the GtkEntry API...
+	 */
+	g_signal_connect_swapped (gtk_entry,
+				  "move-cursor",
+				  G_CALLBACK (move_cursor_cb),
+				  gspell_entry);
+
+	g_signal_connect (GTK_EDITABLE (gtk_entry),
+			  "delete-text",
+			  G_CALLBACK (delete_text_before_cb),
+			  gspell_entry);
+
+	g_signal_connect_swapped (gtk_entry,
+				  "notify::visibility",
+				  G_CALLBACK (emit_changed_signal),
+				  gspell_entry);
+
 	update_buffer (gspell_entry);
 
 	g_object_notify (G_OBJECT (gspell_entry), "entry");
@@ -741,6 +1014,17 @@ gspell_entry_dispose (GObject *object)
 	G_OBJECT_CLASS (gspell_entry_parent_class)->dispose (object);
 }
 
+static void
+gspell_entry_finalize (GObject *object)
+{
+	GspellEntry *gspell_entry = GSPELL_ENTRY (object);
+
+	/* Internal GObject, we can release it in finalize. */
+	g_clear_object (&gspell_entry->current_word_policy);
+
+	G_OBJECT_CLASS (gspell_entry_parent_class)->finalize (object);
+}
+
 static void
 gspell_entry_class_init (GspellEntryClass *klass)
 {
@@ -749,6 +1033,7 @@ gspell_entry_class_init (GspellEntryClass *klass)
 	object_class->get_property = gspell_entry_get_property;
 	object_class->set_property = gspell_entry_set_property;
 	object_class->dispose = gspell_entry_dispose;
+	object_class->finalize = gspell_entry_finalize;
 
 	/**
 	 * GspellEntry:entry:
@@ -772,6 +1057,10 @@ gspell_entry_class_init (GspellEntryClass *klass)
 	 *
 	 * Whether the inline spell checking is enabled.
 	 *
+	 * Even if this property is %TRUE, #GspellEntry disables internally the
+	 * inline spell checking in case the #GtkEntry:visibility property is
+	 * %FALSE.
+	 *
 	 * Since: 1.4
 	 */
 	g_object_class_install_property (object_class,
@@ -787,6 +1076,7 @@ gspell_entry_class_init (GspellEntryClass *klass)
 static void
 gspell_entry_init (GspellEntry *gspell_entry)
 {
+	gspell_entry->current_word_policy = _gspell_current_word_policy_new ();
 }
 
 /**
@@ -899,7 +1189,7 @@ gspell_entry_get_entry (GspellEntry *gspell_entry)
  * gspell_entry_get_inline_spell_checking:
  * @gspell_entry: a #GspellEntry.
  *
- * Returns: whether the inline spell checking is enabled.
+ * Returns: the value of the #GspellEntry:inline-spell-checking property.
  * Since: 1.4
  */
 gboolean
@@ -915,7 +1205,7 @@ gspell_entry_get_inline_spell_checking (GspellEntry *gspell_entry)
  * @gspell_entry: a #GspellEntry.
  * @enable: the new state.
  *
- * Enables or disables the inline spell checking.
+ * Sets the #GspellEntry:inline-spell-checking property.
  *
  * Since: 1.4
  */
