@@ -23,7 +23,7 @@
 /* The same functions as the gtk_text_iter_* equivalents, but take into account:
  * - Word contractions with an apostrophe. For example "doesn't", which is a
  *   contraction of the two words "does not".
- * - Componds with words separated by dashes. For example "spell-checking".
+ * - Compounds with words separated by dashes. For example "spell-checking".
  *
  * When to include an apostrophe or a dash in a word? The heuristic is that the
  * apostrophe must be surrounded by a pango-defined word on *each* side of the
@@ -40,6 +40,9 @@
  * the apostrophe. The implementation would be slightly more complicated, since
  * a function behavior depends on the other side of the word.
  *
+ * When doing changes to the algo here, it should be reflected for the GtkEntry
+ * support as well, to have a consistent behavior.
+ *
  * If the following Pango bug is fixed, the gtk_text_iter_* functions can maybe
  * be used directly.
  * FIXME: https://bugzilla.gnome.org/show_bug.cgi?id=97545
@@ -53,10 +56,7 @@ is_apostrophe_or_dash (const GtkTextIter *iter)
 
 	ch = gtk_text_iter_get_char (iter);
 
-	return (ch == '-' ||
-		ch == '\'' ||
-		ch == _GSPELL_MODIFIER_LETTER_APOSTROPHE ||
-		ch == _GSPELL_RIGHT_SINGLE_QUOTATION_MARK);
+	return _gspell_utils_is_apostrophe_or_dash (ch);
 }
 
 gboolean
